@@ -397,7 +397,7 @@ region is not active.")
 				  conversion-keys simple)
   "Define NAME as a new Quail package for input LANGUAGE.
 TITLE is a string to be displayed at mode-line to indicate this package.
-Optional arguments are GUIDANCE, DOCSTRING, TRANLSATION-KEYS,
+Optional arguments are GUIDANCE, DOCSTRING, TRANSLATION-KEYS,
  FORGET-LAST-SELECTION, DETERMINISTIC, KBD-TRANSLATE, SHOW-LAYOUT,
  CREATE-DECODE-MAP, MAXIMUM-SHORTEST, OVERLAY-PLIST,
  UPDATE-TRANSLATION-FUNCTION, CONVERSION-KEYS and SIMPLE.
@@ -457,7 +457,7 @@ covers Quail translation region.
 
 UPDATE-TRANSLATION-FUNCTION if non-nil is a function to call to update
 the current translation region accoding to a new translation data.  By
-default, a tranlated text or a user's key sequence (if no transltion
+default, a translated text or a user's key sequence (if no translation
 for it) is inserted.
 
 CONVERSION-KEYS specifies additional key bindings used while
@@ -808,9 +808,16 @@ If it is a string, each character is a candidate for the translation.
 If it is a vector, each element (string or character) is a candidate
   for the translation.
 In these cases, a key specific Quail map is generated and assigned to KEY.
+(A Quail map is a cons that satisfies `quail-map-p', which see.)
 
-If TRANSLATION is a Quail map or a function symbol which returns a Quail map,
- it is used to handle KEY."
+If TRANSLATION is a Quail map it is used to handle KEY.  
+
+If TRANSLATION is a function symbol, that function should take two
+arguments, the input string and its length, and it should return a Quail map
+to be used to handle KEY.  For each KEY that translates to the function,
+that function will be called just once, and the result cached by Quail--this
+means that some of the context-dependent processing you may be tempted to
+do, like manipulating unread-command-events, is not possible.  "
   `(quail-install-map
     ',(let ((l rules)
 	    (map (list nil)))
@@ -844,9 +851,17 @@ If it is a cons, the car is one of the above and the cdr is a function
  variable `quail-current-data').  If the cdr part is not a function,
  the value itself is assigned to `quail-current-data'.
 In these cases, a key specific Quail map is generated and assigned to KEY.
+(A Quail map is a cons that satisfies `quail-map-p', which see.)
+
+If TRANSLATION is a Quail map it is used to handle KEY.  
+
+If TRANSLATION is a function symbol, that function should take two
+arguments, the input string and its length.  It should return a Quail map to
+be used to handle KEY.  For each KEY that translates to the function, that
+function will be called just once, and the result cached by Quail--this
+means that some of the context-dependent processing you may be tempted to
+do, like manipulating `unread-command-events', is not possible. 
 
-If TRANSLATION is a Quail map or a function symbol which returns a Quail map,
- it is used to handle KEY.
 Optional argument NAME, if specified, says which Quail package
 to define this translation rule in.  The default is to define it in the
 current Quail package."
@@ -950,7 +965,7 @@ selected translation."
    (t
     (error "Invalid object in Quail map: %s" def))))
 
-(defun quail-lookup-key (key &optional len)
+(defun quail-lookup-key (key &optional len not-reset-indices)
   "Lookup KEY of length LEN in the current Quail map and return the definition.
 The returned value is a Quail map specific to KEY."
   (or len
@@ -988,7 +1003,7 @@ The returned value is a Quail map specific to KEY."
 	  (if (and (consp translation) (vectorp (cdr translation))) 
 	      (progn
 		(setq quail-current-translations translation)
-		(if (quail-forget-last-selection)
+		(if (and (not not-reset-indices) (quail-forget-last-selection))
 		    (setcar (car quail-current-translations) 0))))
 	  ;; We may have to reform cdr part of MAP.
 	  (if (and (cdr map) (symbolp (cdr map)))
@@ -1497,7 +1512,7 @@ or in a newly created frame (if the selected frame has no other windows)."
 	      (quail-make-guidance-frame quail-guidance-buf)
 	    ;; Find the bottom window and split it if necessary.
 	    (let (height)
-	      (setq win (window-at 0 (- (frame-height) 2)))
+	      (setq win (frame-lowest-window))
 	      (setq height (window-height win))
 	      ;; If WIN is tall enough, split it vertically and use
 	      ;; the lower one.
@@ -1635,7 +1650,7 @@ or in a newly created frame (if the selected frame has no other windows)."
 				       (1+ (- idx start)))))
 	      (let ((pos (point)))
 		(insert (aref (cdr quail-current-translations) idx))
-		(if (= idx cur)
+		(if (and (overlayp quail-overlay) (= idx cur))
 		    (move-overlay quail-overlay pos (point))))
 	      (setq idx (1+ idx)))))
       )))
@@ -1839,9 +1854,126 @@ key		binding
 ---		-------\n"))
 	(help-mode)))))
 
+
+;; Quail map generator from state transition table.
+
+(defun quail-map-from-table (table)
+  "Make quail map from state transition table TABLE.
+
+TABLE is an alist, the form is:
+  ((STATE-0 TRANSITION-0-1 TRANSITION-0-2 ...) (STATE-1 ...) ...)
+
+STATE-n are symbols to denote state.  STATE-0 is the initial state.
+
+TRANSITION-n-m are transition rules from STATE-n, and have the form
+\(RULES . STATE-x) or RULES, where STATE-x is one of STATE-n above,
+RULES is a symbol whose value is an alist of keys \(string) vs the
+correponding characters or strings.  The format of the symbol value of
+RULES is the same as arguments to `quail-define-rules'.
+
+If TRANSITION-n-m has the form (RULES . STATE-x), it means that
+STATE-n transits to STATE-x when keys in RULES are input.  Recursive
+transition is allowed, i.e. STATE-x may be STATE-n.
+
+If TRANSITION-n-m has the form RULES, the transition terminates
+when keys in RULES are input.
+
+The generated map can be set for the current Quail package by the
+function `quail-install-map' (which see)."
+  (let ((state-alist (mapcar (lambda (x) (list (car x))) table))
+	tail elt)
+    ;; STATE-ALIST is an alist of states vs the correponding sub Quail
+    ;; map.  It is now initialized to ((STATE-0) (STATE-1) ...).
+    ;; Set key sequence mapping rules in cdr part of each element.
+    (while table
+      (quail-map-from-table-1 state-alist (car table))
+      (setq table (cdr table)))
+
+    ;; Now STATE-ALIST has the form ((STATE-0 MAPPING-RULES) ...).
+    ;; Elements of MAPPING-RULES may have the form (STATE-x).  Replace
+    ;; them with MAPPING-RULES of STATE-x to make elements of
+    ;; STATE-ALIST valid Quail maps.
+    (setq tail state-alist)
+    (while tail
+      (setq elt (car tail) tail (cdr tail))
+      (quail-map-from-table-2 state-alist elt))
+
+    ;; Return the Quail map for the initial state.
+    (car state-alist)))
+
+;; STATE-INFO has the form (STATE TRANSITION ...).  Set key sequence
+;; mapping rules in the element of STATE-ALIST that corresponds to
+;; STATE according to TRANSITION ...
+(defun quail-map-from-table-1 (state-alist state-info)
+  (let* ((state (car state-info))
+	 (map (assq state state-alist))
+	 (transitions (cdr state-info))
+	 elt)
+    (while transitions
+      (setq elt (car transitions) transitions (cdr transitions))
+      (let (rules dst-state key trans)
+	;; ELT has the form (RULES-SYMBOL . STATE-x) or RULES-SYMBOL.
+	;; STATE-x is one of car parts of STATE-ALIST's elements.
+	(if (consp elt)
+	    (setq rules (symbol-value (car elt))
+		  ;; Set (STATE-x) as branches for all keys in RULES.
+		  ;; It is replaced with actual branches for STATE-x
+		  ;; later in `quail-map-from-table-2'.
+		  dst-state (list (cdr elt)))
+	  (setq rules (symbol-value elt)))
+	(while rules
+	  (setq key (car (car rules)) trans (cdr (car rules))
+		rules (cdr rules))
+	  (if (stringp trans)
+	      (if (= (length trans) 1)
+		  (setq trans (aref trans 0))
+		(setq trans (string-to-vector trans))))
+	  (set-nested-alist key trans map nil dst-state))))))
+
+;; ELEMENT is one element of STATE-ALIST.  ELEMENT is a nested alist;
+;; the form is:
+;;	(STATE (CHAR NESTED-ALIST) ...)
+;; NESTED-ALIST is a nested alist; the form is:
+;;	(TRANS (CHAR NESTED-ALIST) ...)
+;; or
+;;	(TRANS (CHAR NESTED-ALIST) ... . (STATE-x))
+;; Here, the task is to replace all occurrences of (STATE-x) with:
+;;	(cdr (assq STATE-x STATE-ALIST))
+
+(defun quail-map-from-table-2 (state-alist element)
+  (let ((prev element)
+	(tail (cdr element))
+	 elt)
+    (while (cdr tail)
+      (setq elt (car tail) prev tail tail (cdr tail))
+      (quail-map-from-table-2 state-alist (cdr elt)))
+    (setq elt (car tail))
+    (if (consp elt)
+	(quail-map-from-table-2 state-alist (cdr elt))
+      (setcdr prev (cdr (assq elt state-alist))))))
+
+;; Concatenate translations for all heading substrings of KEY in the
+;; current Quail map.  Here, `heading substring' means (substring KEY
+;; 0 LEN), where LEN is 1, 2, ... (length KEY).
+(defun quail-lookup-map-and-concat (key)
+  (let* ((len (length key))
+	 (translation-list nil)
+	 map)
+    (while (> len 0)
+      (setq map (quail-lookup-key key len t)
+	    len (1- len))
+      (if map
+	  (let* ((def (quail-map-definition map))
+		 (trans (if (consp def) (aref (cdr def) (car (car def)))
+			  def)))
+	    (if (integerp trans)
+		(setq trans (char-to-string trans)))
+	    (setq translation-list (cons trans translation-list)))))
+    (apply 'concat translation-list)))
+
 
 (defvar quail-directory-name "quail"
-  "Name of Quail directory which cotains Quail packages.
+  "Name of Quail directory which contains Quail packages.
 This is a sub-directory of LEIM directory.")
 
 ;;;###autoload
